# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.


class CreateSocketError(Exception):
    """
    Raised when there is an issue with creating or using the rig's uds socket
    """

    def __init__(self, addr=''):
        message = "Cannot create socket %s" % addr
        super(SocketException, self).__init__(message)


class BindSocketError(Exception):
    """
    Raised when the socket exists, but we cannot communicate with it
    """

    def __init__(self):
        message = "Cannot communicate with rig"
        super(BindSocketError, self).__init__(message)


class MissingSocketError(Exception):
    """
    Raised when the socket we attempted to use does not exist
    """

    def __init__(self, addr=''):
        message = "Socket %s does not exist" % addr
        super(MissingSocketError, self).__init__(message)


class SocketExistsError(Exception):
    """
    Raised when we try to create a socket that already exists
    """

    def __init__(self, addr=''):
        message = "Socket %s already exists on filesystem" % addr
        super(SocketExistsError, self).__init__(message)


class CannotConfigureRigError(Exception):
    """
    Raised when a rig fails to set itself up properly.
    """

    def __init__(self, msg=''):
        message = "Rig setup failed: %s" % msg
        super(CannotConfigureRigError, self).__init__(message)


class DestroyRig(Exception):
    """
    Raised when we intentionally destroy a rig, so we can trap the exit of the
    thread pool
    """


__all__ = [
    'BindSocketError',
    'CannotConfigureRigError',
    'CreateSocketError',
    'DestroyRig',
    'MissingSocketError',
    'SocketExistsError'
]



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import logging
import os
import shlex
import shutil
import time

from rigging.exceptions import *
from subprocess import Popen, PIPE


class BaseAction():
    """
    Base class that all rig actions need to subclass.

    Actions are what rig does when a resource monitor (the rig) is triggered.
    This will typically be a one-off execution of some command, for example
    generating an application core when a log message is matched. In this
    example the log message is matched by the Logs rig, and the generation of
    the application core would be the Action.
    """

    enabling_opt = ''
    enabling_opt_desc = ''
    action_name = ''
    priority = 10
    required_binaries = ()
    repeatable = False

    def __init__(self, rig):
        self.rig = rig
        self.id = self.rig.id
        self.repeat_count = 0

    def set_tmp_dir(self, tmp_dir):
        if os.path.exists(tmp_dir):
            self.tmp_dir = tmp_dir
        else:
            raise CannotConfigureRigError("%s does not exist" % tmp_dir)

    def load(self, args):
        """
        Actually load the commandline configuration into the instantiated
        action.
        """
        self.args = args
        self.detached = False
        self.report_files = []
        self.report_message = None
        self._setup_action_logging()
        loaded = self._pre_action()
        if not loaded:
            self.log_error("Failed to load action %s"
                           % self.action_name)
            self.cleanup()
        return loaded

    def _setup_action_logging(self):
        extra = {'rig_id': self.id}
        self.logger = logging.getLogger('rig')
        self.logger = logging.LoggerAdapter(self.logger, extra)
        self.console = logging.getLogger('rig_console')
        self.console = logging.LoggerAdapter(self.console, extra)

    def log_error(self, msg):
        self.logger.error(msg)
        if not self.detached:
            self.console.error(msg)

    def log_info(self, msg):
        self.logger.info(msg)
        if not self.detached:
            self.console.info(msg)

    def log_debug(self, msg):
        self.logger.debug(msg)
        if not self.detached:
            self.console.debug(msg)

    def action_info(self):
        """
        Return information about what the action should generate. This is
        called during 'rig info <rig id>' and appears as the 'expected_result'
        element.

        MUST be overridden by specific actions
        """
        return 'Not defined by action'

    def get_option(self, option):
        """
        Return the value for the given option from the rig or the command line
        args
        """
        return self.rig.get_option(option)

    def exec_cmd(self, cmd):
        """
        Executes the given command via Popen() without getting a TTY.

        Positional arguments
            cmd - a string containing the full command to run

        Returns
            dict containing the exit status and output from the cmd
        """
        self.log_debug("Running command %s" % cmd)
        cmd = shlex.split(cmd)
        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, encoding='utf-8',
                     shell=False)
        stdout, stderr = proc.communicate(timeout=180)
        rc = proc.returncode
        return {'status': rc, 'stdout': stdout, 'stderr': stderr}

    @classmethod
    def _add_action_options(cls, parser):
        """Add an argument group for the action to better format --help output
        """
        act_grp = parser.add_argument_group(
            "%s Action Options" % cls.__name__.title()
        )
        cls.add_action_options(act_grp)
        return parser

    @classmethod
    def add_action_options(cls, parser):
        """
        This is where the action-specific options are added.

        Returns:
            parser - ArgumentParser (sub) parser object
        """
        pass

    def _check_exists(self, binary):
        """
        Checks to see if the given binary exists in PATH.
        """
        paths = os.environ.get("PATH", "").split(os.path.pathsep)
        cmds = [os.path.join(p, binary) for p in paths]
        return any(os.access(path, os.X_OK) for path in cmds)

    def _pre_action(self):
        """
        This is called prior to setting up a resource monitor, and wraps an
        Action's pre_action() method, if one is defined.

        Returns:
            bool - True if pre_action completes successfully or if no
                   pre_action is defined. Raises an exception if not successful
        """
        try:
            for binary in self.required_binaries:
                if not self._check_exists(binary):
                    self.log_error("Required binary %s not found" % binary)
                    return False
            return self.pre_action()
        except CannotConfigureRigError:
            raise
        except Exception as err:
            self.log_error("Could not execute pre-action for action %s: %s"
                           % (self.action_name, err))
            return False

    def pre_action(self):
        """
        MAY be overriden by any rig action subclassing BaseAction.

        Prior to a monitor starting, this is called to allow rig actions to
        perform any needed setup.

        For example the network action may start a packet capture during rig
        creation and uses a pre-action to do so. The trigger_action would then
        be stopping the packet capture.
        """
        return True

    def _trigger_action(self):
        """
        This is called whenever an action should be triggered by a monitor on
        a resource.

        The action's trigger_action() method must override the default
        trigger_action() method.

        Returns:
            bool - True if trigger completes without issue or an exception
        """
        try:
            self.trigger_action()
            if self.get_option('repeat') > 0 and self.repeatable:
                while self.repeat_count < self.get_option('repeat'):
                    # sleep here instead of after the trigger so that the first
                    # repeat is also delayed from the initial execution.
                    time.sleep(self.get_option('repeat_delay'))
                    self.repeat_count += 1
                    self.log_info("Triggering action %s again. Repeat count is"
                                  " now %s. Will repeat %s times total"
                                  % (self.action_name, self.repeat_count,
                                     self.get_option('repeat')))
                    self.trigger_action()
            return True
        except Exception as err:
            self.log_error("Exception triggering action %s: %s" %
                           (self.action_name, err))
            raise

    def trigger_action(self):
        """
        MUST be overriden by a rig action subclassing BaseAction.

        This method should perform all needed actions for when a resource
        monitor is triggered.

        Does not need to return any specific value, but should raise exceptions
        when the action cannot be completed successfully (rather than returning
        False)

        If an action is generating a file to be collected, it needs to be
        handled by the action's report_result() rather than here.
        """
        raise NotImplementedError

    def _post_action(self):
        """
        This is called after a resource monitor has triggered, the action's
        trigger_action() has been called, and the resource monitor is no longer
        running.

        Returns:
            bool - True if post action completes, or raises an exception
        """
        try:
            self.post_action()
            return True
        except Exception:
            raise

    def post_action(self):
        """
        MAY be overridden by a rig action subclassing BaseAction.

        This method should perform any necessary cleanup as a result of the
        action's trigger_action().

        Does not need to return any specific value, but should raise exceptions
        when the post action cannot be comepleted successfully.
        """
        pass

    def finish_execution(self):
        """
        Called when an aciton has completed it's trigger, and we need to
        perform closing operations to cleanly exit the action.
        """
        self._report_results()
        return self.report_files
        # TODO: consider tar'ing up the tmp directory

    def _report_results(self):
        """
        This is called at the end of execution.

        This will create a single log entry that includes any of the files
        the action created and used add_report_files() with, as well as the
        add_report_message() string(s).
        """
        if not self.report_files and not self.report_message:
            return
        msg = "Action %s" % self.action_name
        if self.report_files:
            self.log_debug(msg + " generated the following files: %s" %
                           ','.join(f for f in self.report_files))
        if self.report_message:
            self.log_info(msg + " generated the following message: %s" %
                          (self.report_message))

    def add_report_file(self, filename):
        """
        Used to report file(s) that have been generated to the user.

        Positional arguments:
            filename (list or str) - The full path of the file to report to the
                    user. Can accept a string or list of strings
        """
        if not isinstance(filename, list):
            filename = [filename]
        for fname in filename:
            if not fname.startswith(self.tmp_dir):
                try:
                    shutil.move(fname, self.tmp_dir)
                except OSError:
                    self.log_error("Could not move %s to rig tmp dir %s"
                                   % (fname, self.tmp_dir))
        self.report_files.extend(filename)

    def add_report_message(self, message):
        """
        Used to provide arbritary text notifications to the user.

        Note that in the event the rig calling this action is run daemonized,
        any messages will only appear in the rig log

        Positional arguments:
            message (str) - The message to be printed/reported to the user.
        """
        if not isinstance(message, str):
            raise TypeError('message must be of type str')
        self.report_message = message

    def cleanup(self):
        """
        Used on select actions that may have bits that need to be cleanup up
        in the event that a rig fails to start, but are not automatically
        handled by the system.

        For example, with the tcpdump action there is a possiblity we could
        fail to start the rig, but still have a detached thread running the
        tcpdump command that we would want to kill
        """
        pass



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import inspect
import json
import logging
import os
import random
import shutil
import string
import socket
import sys
import tarfile
import time

from argparse import Action
from concurrent.futures import ThreadPoolExecutor, wait, FIRST_COMPLETED
from concurrent.futures import thread
from datetime import datetime
from rigging.actions import BaseAction
from rigging.exceptions import *

RIG_DIR = '/var/run/rig/'
RIG_TMP_DIR_PREFIX = '/var/tmp/rig'


class RigExtendOption(Action):
    """
    Implementation of the 'extend' action from python-3.8+ for older runtimes

    This will be registered as 'rigextend' to not clobber newer runtimes.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        items = [opt for opt in values.split(',')]
        if getattr(namespace, self.dest):
            items += getattr(namespace, self.dest)
        setattr(namespace, self.dest, items)


class BaseRig():
    """
    Base class for resources/triggers.

    Will build a set of argparse options based upon what the rig provides by
    itself, as well as the defined actions it supports.

    Two class members must be defined by a rig:
        resource_name: The canonical name of the rig as used to invoke it
                       on the commandline.
        parser_description: The argparse description text for the Rig's submenu
        parser_usage:  The argparse usage text for the Rig's submenu
    """

    parser_description = None
    parser_usage = """
    rig %(name)s <options>

    Valid actions:

    """
    triggered = False
    watcher_threads = []
    rig_wide_opts = ()
    _triggered_from_cmdline = False

    def __init__(self, parser):
        self.detached = False
        self._status = 'Initializing'
        self.resource_name = self.__class__.__name__.lower()
        self.parser_usage = self.parser_usage % {'name': self.resource_name}
        self.pool = None
        self.parser = parser
        self.restart_count = 0
        subparser = self.parser.add_subparsers()
        self.rig_parser = subparser.add_parser(self.resource_name)
        self.rig_parser.register('action', 'rigextend', RigExtendOption)
        self.rig_parser = self._setup_parser(self.rig_parser)

        self.supported_actions = self._load_supported_actions()
        for action in self.supported_actions:
            self.supported_actions[action]._add_action_options(self.rig_parser)
            self.parser_usage += '\t{:<15} {:>30}\n'.format(
                self.supported_actions[action].enabling_opt,
                self.supported_actions[action].enabling_opt_desc
            )

        if self.parser_description:
            self.rig_parser.description = self.parser_description
        if self.parser_usage:
            self.rig_parser.usage = self.parser_usage

        self._can_run = self._load_args()
        if self._can_run:
            self.set_rig_id()
            self.pid = os.getpid()
            self.created_time = datetime.strftime(datetime.now(),
                                                  '%D %H:%M:%S')
            self.rig_options = {}
            self._load_rig_wide_options()
            self._setup_rig_logging()
            self.log_debug("Initializing %s rig %s" %
                           (self.resource_name, self.id))
            self._sock, self._sock_address = self._create_rig_socket()
            self._tmp_dir = self._create_temp_dir()
            self.files = []

    def set_rig_id(self):
        """If the --name option is given, update the rig's ID to that value.
        """
        if self.args['name']:
            self.id = self.args['name']
        else:
            self.id = (''.join(random.choice(string.ascii_lowercase)
                       for x in range(5)))

    def _exit(self, errno):
        """
        Handles pre-mature exits due to errors
        """
        self._cleanup_threads()
        self._cleanup_socket()
        raise SystemExit(errno)

    def _detach(self):
        """
        Here we effectively daemonize the process by using the double-fork
        method. The rig will continue to run until a trigger event, or until
        the rig cli is used to send a termination signal to the socket the rig
        is listening on.
        """
        def _fork():
            try:
                pid = os.fork()
                if pid > 0:
                    sys.exit(0)
            except OSError as e:
                self.log_error("Fork failed: %s" % e)
                sys.exit(1)
        _fork()
        os.setsid()
        os.umask(0)
        _fork()

        sys.stdin = sys.__stdin__
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        sys.stdout.flush()
        sys.stderr.flush()
        _std = getattr(os, 'devnull', '/dev/null')
        sin = open(_std, 'r')
        sout = open(_std, 'a+')
        serr = open(_std, 'a+')
        os.dup2(sin.fileno(), sys.stdin.fileno())
        os.dup2(sout.fileno(), sys.stdin.fileno())
        os.dup2(serr.fileno(), sys.stderr.fileno())

        self.pid = os.getpid()
        return True

    def _create_rig_socket(self):
        """
        Creates the UNIX socket that the rig will listen on for lifecycle
        management.

        This socket is used by the rig cli when getting status information or
        destroying a deployed rig before the trigger event happens.
        """
        if not os.path.exists(RIG_DIR):
            os.makedirs(RIG_DIR)
        _sock_address = "%s%s" % (RIG_DIR, self.id)
        try:
            os.unlink(_sock_address)
        except OSError:
            if os.path.exists(_sock_address):
                raise SocketExistsError(_sock_address)
        try:
            _sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            _sock.bind(_sock_address)
            _sock.listen(1)
            return (_sock, _sock_address)
        except Exception as err:
            self.log_error("Unable to create unix socket: %s" % err)
            raise CreateSocketException

    def _create_temp_dir(self):
        """
        Create a temp directory for rig to use for saving created files too
        """
        try:
            _dir = "%s.%s/" % (RIG_TMP_DIR_PREFIX, self.id)
            os.makedirs(_dir)
            return _dir
        except OSError:
            raise CannotConfigureRigError('failed to create temp directory')

    def _load_args(self):
        """
        Parses the args given to us by the user.

        This is called while trying to load a rig. If we do not have any args,
        then that means that '--help' was called, in which case we return False
        to ensure we don't begin logging for no reason.

        If there is an unknown option provided, argparse appends it to another
        namespace list, thus if this list contains more than just the resource
        as an element, it means we have an unknown arg.
        """
        args = self.rig_parser.parse_known_args()
        filt = ['--debug', '--foreground']
        unknowns = [x for x in args[1][1:] if x not in filt]
        if len(unknowns):
            print("Unknown option %s specified." %
                  unknowns[0].split('=')[0])
            return False
        self.args = vars(self.rig_parser.parse_known_args()[0])
        self.debug = self.args['debug']
        if self.args:
            return True
        return False

    def _load_supported_actions(self):
        """
        Looks at the defined actions available to rig, and if they match any
        of the strings listed in supported_actions instantiates them.
        """
        actions = {}
        import rigging.actions
        pkg = rigging.actions
        for path in pkg.__path__:
            if os.path.isdir(path):
                for pyfile in sorted(os.listdir(path)):
                    if not pyfile.endswith('.py') or '__' in pyfile:
                        continue
                    fname, ext = os.path.splitext(pyfile)
                    modname = 'rigging.actions.%s' % fname
                    mod_short_name = modname.split('.')[2]
                    mod = __import__(modname, globals(), locals(),
                                     [mod_short_name])
                    modules = inspect.getmembers(mod, inspect.isclass)
                    for module in modules:
                        if module[1] == BaseAction:
                            continue
                        if not issubclass(module[1], BaseAction):
                            continue
                        actions[module[1].action_name] = module[1]
        return actions

    def _load_rig_wide_options(self):
        """
        Based on the rig's rig_wide_options member, take the values for those
        options and load them into the rig_options dict to be used for global
        usage in actions, so we can avoid having to specify the same value over
        and over between rigs and actions.
        """
        for opt in self.rig_wide_opts:
            if opt in self.args:
                self.rig_options[opt] = self.args[opt]

    def _setup_parser(self, parser):
        """
        Builds the option parser based on supported actions, then appends the
        rig-specific options.

        Returns:
            parser: an ArgumentParser object that contains the rig-specific
                    options.
        """
        # Add the common rigging options here
        global_grp = parser.add_argument_group('Global Options')
        global_grp.add_argument('--foreground', action='store_true',
                                default=False,
                                help='Run the rig in the foreground')
        global_grp.add_argument('--debug', action='store_true',
                                help='Print debug messages to console')
        global_grp.add_argument('--delay', type=int, default=0,
                                help='Seconds to delay running actions')
        global_grp.add_argument('--name', type=str, default='',
                                help='Specify a name for the rig')
        global_grp.add_argument('--no-archive', default=False,
                                action='store_true',
                                help='Do not create a tar archive of results')
        global_grp.add_argument('--restart', default=0, type=int,
                                help='Number of times a rig should restart '
                                     'itself')
        global_grp.add_argument('--repeat', default=0, type=int,
                                help=('Number of times to repeat actions that '
                                      'support repitition'))
        global_grp.add_argument('--repeat-delay', default=1, type=int,
                                help='Seconds to delay between repeating '
                                     'actions')
        global_grp.add_argument('--interval', default=1, type=int,
                                help='Time to wait between rig polling '
                                     'intervals')
        return self._set_parser_options(parser)

    def compile_details(self):
        try:
            args = sys.argv[2:]
            return ' '.join(args)[:40]
        except Exception:
            return ''

    @property
    def watching(self):
        """
        MUST be overridden by rigs. This should return a string describing
        what resource(s) the rig is monitoring
        """
        return NotImplementedError

    @property
    def trigger(self):
        """
        MUST be overridden by rigs. This should return a string containing the
        trigger event for the monitored resource.
        """
        return NotImplementedError

    @property
    def status(self, value=None):
        """
        Returns the current status of the rig.

        :param value: Unused, but passed by the listening socket
        """
        return {
            'id': self.id,
            'pid': str(self.pid),
            'rig_type': self.resource_name,
            'watch': self.watching[:30],
            'trigger': self.trigger[:35],
            'status': self._status
        }

    @property
    def info(self, value=None):
        """
        Returns detailed information about the rig for a more in-depth view
        than status provides
        """
        return {
            'id': self.id,
            'pid': str(self.pid),
            'rig_type': self.resource_name,
            'status': self._status,
            'restart_max': self.args['restart'],
            'restart_count': self.restart_count,
            'cmdline': " ".join(sys.argv),
            'debug': self.debug,
            'watch': self.watching,
            'trigger': self.trigger,
            'created': self.created_time,
            'actions': self._get_action_info()
        }

    def _setup_rig_logging(self):
        extra = {'rig_id': self.id}
        self.logger = logging.getLogger('rig')
        self.logger = logging.LoggerAdapter(self.logger, extra)
        self.console = logging.getLogger('rig_console')
        self.console = logging.LoggerAdapter(self.console, extra)
        if self.get_option('debug'):
            self.logger.setLevel(logging.DEBUG)
            self.console.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

    def log_error(self, msg):
        self.logger.error(msg)
        if not self.detached:
            self.console.error(msg)

    def log_info(self, msg):
        self.logger.info(msg)
        if not self.detached:
            self.console.info(msg)

    def log_debug(self, msg):
        self.logger.debug(msg)
        if not self.detached and self.debug:
            self.console.debug(msg)

    def log_warn(self, msg):
        self.logger.warn(msg)
        if not self.detached:
            self.console.warn(msg)

    def set_option(self, option, value):
        """
        Override the rig_wide_option for OPTION with VALUE.
        """
        self.rig_options[option] = value

    def get_option(self, option):
        """
        Retrieve a specified option from the loaded commandline options.

        An invalid option returns as False, rather than raises an exception.

        Returns
            str or bool - If the option has a value other than True, it is
                returned as a string. Otherwise return True or False depending
                on if it has a value at all.
        """
        if option in self.rig_options.keys() and self.rig_options[option]:
            return self.rig_options[option]
        if option in self.args.keys():
            _opt = self.args[option]
            # if the option is not set from the cmdline, and it is loaded from
            # the rig options (as opposed to action options), give the rig-wide
            # option
            if not _opt and option in self.rig_options.keys():
                return self.rig_options[option]
            # otherwise, provide the action-specific option value
            else:
                return _opt
        return False

    def _set_parser_options(self, parser):
        """Internal helper that automatically creates a new argument group
        for each rig
        """
        rig_group = parser.add_argument_group(
            "%s Rig Options" % self.__class__.__name__
        )
        self.set_parser_options(rig_group)
        return parser

    def set_parser_options(self, parser):
        """
        This is where the rig-specific options are actually specified.

        Returns:
            parser - ArgumentParser (sub) parser object
        """
        pass

    def _fmt_return(self, command, output='', success=True):
        """
        Formats a return value as a dict specifying the id of this rig, the
        command run, any output, and if the command was successful
        """
        return json.dumps({
            'id': self.id,
            'command': command,
            'result': output,
            'success': success
        }).encode()

    def _listen_on_socket(self):
        self.log_debug('Listening on %s' % self._sock_address)
        while True:
            conn, client = self._sock.accept()
            try:
                req = json.loads(conn.recv(1024).decode())
            except Exception as err:
                self.log_debug("Error parsing socket request: %s" % err)
                return self._fmt_return(command="Unknown", result=err,
                                        success=False)
            self.log_debug("Received request '%s' from socket"
                           % req['command'])
            if req['command'] == 'destroy':
                self._status = 'destroying'
                self.log_debug("Shutting down rig")
                ret = self._fmt_return(command='destroy')
                conn.sendall(ret)
                raise DestroyRig
            try:
                ret = str(getattr(self, req['command']))
                self.log_debug("Sending '%s' back to client" % ret)
                conn.sendall(self._fmt_return(command=req['command'],
                                              output=ret))
            except Exception as err:
                self.log_debug(err)
                self.log_error('No attribute: %s' % req['command'])
                conn.sendall(self._fmt_return(command=req['command'],
                                              output='No such attribute',
                                              success=False))
            continue

    def _register_actions(self):
        """
        Compare the commandline options to supported actions for the rig.

        For any options matched against the supported actions, we initialize
        those actions to then be triggered once the rig hits the triggering
        conditions.
        """
        self._actions = {}
        for action in self.supported_actions:
            _act = self.supported_actions[action]
            if action in self.args and self.args[_act.enabling_opt]:
                _action = self.supported_actions[action](self)
                _action.set_tmp_dir(self._tmp_dir)
                loaded = _action.load(self.args)
                if not loaded:
                    self._exit(1)
                self._actions[action] = _action

    def _get_action_info(self):
        """
        Provide detailed information about the actions that will be taken
        when the rig is triggered

        Returns:
            dict of action dicts with associated information
        """
        acts = {}
        for action in self._actions:
            _act = self._actions[action]
            acts[action] = {
                'name': action,
                'priority': _act.priority,
                'expected_result': _act.action_info()
            }
        return acts

    def setup(self):
        """
        MUST be overridden by rigs subclassing BaseRig.

        This is where rigs will define their watcher threads.
        """
        raise NotImplementedError

    def execute(self):
        """
        Main entry point for rigs.
        """
        try:
            # detach from console
            if not self.args['foreground']:
                print(self.id)
                self._detach()
                self.detached = True
            self.setup()
            self._register_actions()
            if self.detached:
                for action in self._actions:
                    self._actions[action].detached = True
            ret = self._create_and_monitor()
            if self.args['restart'] != 0:
                while (self.restart_count < self.args['restart'] or
                        self.args['restart'] == -1):
                    self.restart_count += 1
                    self.log_info("Restarting rig %s. Current restart count is"
                                  " %s" % (self.id, self.restart_count))
                    # Re-create the temp dir for the rig every restart so we
                    # don't overlap data being archived
                    self._create_temp_dir()
                    # clear and re-load the configured rig options
                    self.reset_counters()
                    ret = self._create_and_monitor()
            self.trigger_kdump()
            self._status = 'Exiting'
            self._cleanup_socket()
            if ret:
                os._exit(0)
            else:
                os._exit(ret)
        except KeyboardInterrupt:
            self.log_debug('Received keyboard interrupt, destroying rig.')
            self._exit(140)
        except Exception as err:
            self.log_error(err)
            self._exit(1)

    def _create_and_monitor(self):
        """
        Create the threads for listening on the socket and monitoring the bits
        the rig is meant to monitor.

        Blocks until either the monitoring thread returns, or there is a socket
        related problem that interrupts the listening thread.

        After waiting for the first thread to exit, this will clean itself up
        by removing both the thread pool and the temp dir used after the
        results are archived.

        This method may be called multiple times if the rig is configured to
        restart itself after being triggered, hence it is responsible for the
        entire creation and take down process.
        """
        _threads = []
        # listen on the UDS socket in one thread, spin the watcher
        # off in a separate thread
        self._control_pool = ThreadPoolExecutor(2)
        _threads.append(self._control_pool.submit(self._listen_on_socket))
        _threads.append(self._control_pool.submit(self._monitor_resource))
        self._status = 'Running'
        ret = wait(_threads, return_when=FIRST_COMPLETED)
        self.archive_name = self.create_archive()
        self.report_created_files()
        self._cleanup_threads()
        return ret

    def _monitor_resource(self):
        """
        This is the main function in which we watch for a resource's trigger
        condition(s).

        This will block until the rig has self.triggered become True.
        """
        try:
            ret = self.start_watcher_threads()
            if ret:
                self.log_info(
                    'Watcher thread triggered. Stopping other watcher threads')
                self.pool._threads.clear()
                if self.args['delay']:
                    self.log_debug('Delaying trigger for %s seconds'
                                   % self.args['delay'])
                    time.sleep(self.args['delay'])
                self.trigger_actions()
        except Exception:
            raise

    @property
    def manual_trigger(self):
        """
        Manually triggers the rig, kicking off the watcher thread
        """
        self.log_info('Received request to manually trigger rig.')
        self._triggered_from_cmdline = True

    def _watch_for_manual_trigger(self):
        """
        This thread will watch for a manual trigger request from the cmdline,
        and return True iff that request is made
        """
        while not self._triggered_from_cmdline:
            time.sleep(1)
        self.log_debug('Trigger from cmdline received. Triggering watcher')
        return True

    def reset_counters(self):
        """
        Reset whatever counters a rig uses to determine trigger state. This is
        called when a rig restarts itself, so that a previous run does not
        influence the run of the now-restarted rig.

        This SHOULD BE overridden by specific rigs, though if there is not some
        form of counter that the rig uses, it can be ignored.
        """
        pass

    def create_archive(self):
        """
        Takes the contents on the temp directory used for the rig and creates
        a tarball of them, placing the archive in /var/tmp.

        Later, the rig will remove the temp directory for itself.
        """
        if self.get_option('no_archive'):
            self.log_info('Not creating a tar archive of collected data')
            return
        if not os.listdir(self._tmp_dir):
            self.log_info('No data generated to archive for this rig.')
            return
        _arc_date = datetime.strftime(datetime.now(), '%Y-%m-%d-%H%M%S')
        _arc_name = "rig-%s-%s" % (self.id, _arc_date)
        _arc_fname = "/var/tmp/%s.tar.gz" % _arc_name
        with tarfile.open(_arc_fname, 'w:gz') as tar:
            tar.add(self._tmp_dir, arcname=_arc_name)
        return _arc_fname

    def report_created_files(self):
        """
        Report all files created by all actions
        """
        if not self.archive_name and self.files:
            self.log_info("The following files were created for this rig: %s"
                          % ', '.join(f for f in self.files))
        if self.archive_name:
            self.log_info("An archive containing this rig's data is available "
                          "at %s" % self.archive_name)

    def add_watcher_thread(self, target, args):
        """
        Used by rigs to define new thread(s) to start in order to monitor their
        respective resources. Each required thread should make a separate call
        to add_watcher_thread().

        Positional Arguments:
            target - A callable method, almost always defined by the rig
            args - Args that should be passed to the target method, if multiple
                pass this as a tuple.
        """
        if not callable(target):
            raise Exception("Unable to add watcher thread. Target must be a "
                            "callable method, received %s" % target.__class__)
        if not isinstance(args, tuple):
            args = (args, )
        self.watcher_threads.append((target, args))

    def start_watcher_threads(self):
        """
        Start the threadpool and submits the requested watcher threads as jobs.

        Blocks until one of the threads returns, acting as a trigger event for
        the rig
        """
        try:
            futures = []
            self.pool = ThreadPoolExecutor()
            for wthread in self.watcher_threads:
                futures.append(self.pool.submit(wthread[0], *wthread[1]))
            futures.append(self.pool.submit(self._watch_for_manual_trigger))
            results = wait(futures, return_when=FIRST_COMPLETED)
            result = list(results[0])[0].result()
            return result
        except Exception as err:
            self.log_error("Exception caught for rig %s: %s" % (self.id, err))
            self._exit(1)

    def trigger_actions(self):
        """
        This is called when a rig's monitoring condition is met. This will then
        invoke any and all actions defined by the user.
        """
        self._status = 'Triggered'
        self.files = []
        try:
            for action in sorted(self._actions,
                                 key=lambda x: self._actions[x].priority):
                if action == 'kdump':
                    continue
                self.log_debug("Triggering action %s" % action)
                self._actions[action]._trigger_action()
                self.files.extend(self._actions[action].finish_execution())
        except Exception as err:
            self.log_error("Error executing actions: %s" % err)

    def wait_loop(self):
        """Used to sleep a watcher thread for the length of time specified by
        --interval.

        Using this allows a standardized way to both ensure rigs wait the
        amount of time requested by the user between polling intervals, and
        to avoid doing repetitive 'import sleep; time.sleep()' calls.
        """
        time.sleep(self.get_option('interval'))

    def trigger_kdump(self):
        """
        If configured, kdump needs to be triggered at the end of execution.

        Thus, this is called after everything else.
        """
        if 'kdump' in self._actions.keys():
            # if we don't remove this here, we'll have a stale socket for
            # every kdump rig created
            self._cleanup_socket()
            self._actions['kdump']._trigger_action()

    def _cleanup_threads(self):
        try:
            for action in self._actions:
                self._actions[action].cleanup()
            self.pool.shutdown(wait=False)
            self.pool._threads.clear()
            self._control_pool.shutdown(wait=False)
            self._control_pool._threads.clear()
            thread._threads_queues.clear()
        except AttributeError:
            # _exit() called before rig was initialized
            pass

        try:
            if self.archive_name or self._status == 'destroying':
                shutil.rmtree(self._tmp_dir)
        except Exception as err:
            self.log_error("Could not remove temp dir: %s" % err)

    def _cleanup_socket(self):
        try:
            os.remove(self._sock_address)
        except OSError as err:
            if err.errno == 2:
                pass
            else:
                self.log_error("Failed to remove listening socket %s: %s" %
                               (self._sock_address, err))



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from rigging.rigs import BaseRig
from rigging.exceptions import CannotConfigureRigError
from pathlib import Path

import os


class Filesystem(BaseRig):
    """Monitor a filesystem, directory, or file.

    May watch an entity for changes, reaching a size threshold, creation or
    deletion, and more.
    """

    parser_description = 'Monitor a filesystem, directory, or file.'

    def set_parser_options(self, parser):
        parser.add_argument('--path', type=str,
                            help='Specify the path to monitor')

        parser.add_argument('--size', type=str,
                            help=('Trigger if directory/file size exceeds '
                                  'this threshold')
                            )
        parser.add_argument('--fs-used', type=int,
                            help=('Trigger if FS backing --path if over this '
                                  '%% used')
                            )
        parser.add_argument('--fs-size', type=str,
                            help=('Trigger if FS backing --path is over this '
                                  'total size'))
        return parser

    @property
    def watching(self):
        return self.get_option('path')

    @property
    def trigger(self):
        msg = "%s over" % self.get_option('path')
        if self.get_option('size'):
            msg += " %s in size" % self.get_option('size')
        if self.get_option('fs_used'):
            msg += " %s%% used filesystem space" % self.get_option('fs_used')
        if self.get_option('fs_size'):
            msg += " %s filesystem space utilized" % self.get_option('fs_size')
        return msg

    def _fmt_size(self, val):
        """Takes the size value and parses it into a bytes value
        """
        units = {
            'K': 1024,
            'M': 1048576,
            'G': 1073741824,
            'T': 1099511627776
        }
        size = val[:-1]
        unit = val[-1]

        try:
            size = float(size)
        except Exception:
            raise CannotConfigureRigError("Invalid size '%s' provided" % size)

        if unit not in units.keys():
            raise CannotConfigureRigError("Unknown unit '%s' provided" % unit)

        return size * units[unit]

    def setup(self):
        if not self.get_option('path'):
            raise CannotConfigureRigError('Must specify path to watch')

        path = Path(self.get_option('path'))
        if not path.exists():
            raise CannotConfigureRigError(
                "Path '%s' does not exist" % self.get_option('path')
            )

        if self.get_option('size'):
            size = self._fmt_size(self.get_option('size'))
            self.add_watcher_thread(target=self.watch_path_size,
                                    args=(path, size))

        if self.get_option('fs_used') or self.get_option('fs_size'):
            self.add_watcher_thread(target=self.watch_fs_used,
                                    args=path)

    def watch_fs_used(self, path):
        """Watch the filesystem backing the specified path, and trigger if it
        exceeds the --fs-used/size used space
        """
        fs_stat = os.statvfs(path)
        # Get full size of fs in bytes
        fs_size = fs_stat.f_frsize * fs_stat.f_blocks
        # Determine max allowed bytes used based on --fs-used or --fs-size
        if self.get_option('fs_used'):
            fs_max_used = (fs_size * (self.get_option('fs_used') / 100))
        else:
            fs_max_used = self._fmt_size(self.get_option('fs_size'))

        self.log_debug("Determined max allowed used space for %s to be %sB"
                       % (path, fs_max_used))

        while True:
            fs = os.statvfs(path)
            # Get current used amount in bytes
            free = fs.f_frsize * fs.f_bfree
            current_used = fs_size - free
            if current_used > fs_max_used:
                perc = round((current_used/fs_size * 100))
                self.log_info("Used space on %s is %sB (%s%%) exceeding "
                              "threshold of %sB."
                              % (path, current_used, perc, fs_max_used))
                return True
            self.wait_loop()

    def watch_path_size(self, path, size):
        """Watch the specified path and trigger if its size exceeds the
        provided --size in bytes
        """

        while True:
            if path.is_dir():
                cur_size = sum(f.stat().st_size for f in path.glob('**/*')
                               if f.is_file())
            else:
                cur_size = path.stat().st_size

            if cur_size > size:
                self.log_info(
                    "Size of path '%s' is %s bytes, exceeding threshold of %s."
                    % (self.get_option('path'), cur_size, size)
                )
                return True

            self.wait_loop()



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import os
import re
import select

from rigging.rigs import BaseRig
from rigging.exceptions import CannotConfigureRigError
from systemd import journal


class Logs(BaseRig):
    """
    Supports watching one or more log files and/or one or more journals.
    May also watch the system journal.

    Resource triggers and argparse options:
        :opt logfile: The full path of the file(s) to watch. Defaults to
            /var/log/messages.
        :opt journal: The name of the journal unit to watch. Defaults to
            the system journal.
        :opt no-files: Disable watching log files
        :opt no-journal: Disable watching the journal
        :opt message: Trigger string (capable of shell-style regex) to
            watch logfile or journal for.
    """

    parser_description = ('Watch one or more log files and/or journals for a '
                          'specified log message')

    def set_parser_options(self, subparser):
        subparser.add_argument('--logfile', default='/var/log/messages',
                               help='log(s) to be watched')
        subparser.add_argument('--journal', default='system',
                               help='journal unit(s) to be watched')
        subparser.add_argument('--no-files', default=False,
                               action='store_true',
                               help='Do not watch any log files')
        subparser.add_argument('--no-journal', default=False,
                               action='store_true',
                               help='Do not watch any journals')
        subparser.add_argument('-m', '--message', required=True,
                               help='String to trigger against')
        subparser.add_argument('--count', default=1, type=int,
                               help=('Trigger only after message has been '
                                     'matched this many times'))
        return subparser

    @property
    def watching(self):
        files = (self.get_option('logfile').replace('/var/log/', '') if not
                 self.get_option('no_files') else '')
        units = ('journals: ' + self.get_option('journal') if not
                 self.get_option('no_journal') else 'no journals')
        watch = [files] + [units]
        ret = ', '.join([w for w in watch if w])
        return ret

    @property
    def trigger(self):
        return self.get_option('message')

    def reset_counters(self):
        self.counter = 0

    def _sanitize_message(self, message):
        """
        Inspects the message option given to the rig and will attempt to
        check it for common errors and warn when one is found.
        """
        errmsg = ''
        try:
            # attempt to perform a regex match on itself. This is to catch
            # regex syntax errors early on. Note that we have to strip any of
            # the otherwise required escaping in order to properly self-test
            test_message = message
            for c in '[]()':
                test_message = test_message.replace('\\' + c, c)
            _check = re.match(message, test_message)
            if _check is None:
                # the regex style is syntactically correct for parsing
                # however the message did not trigger on itself, so there must
                # be an error. This usually happens when regex syntax appears
                # in the message directly, like [] or () characters.
                errmsg = ('Message failed self-matching test, check any regex '
                          'provided. Hint: are brackets and parentheses '
                          'escaped?')
        except Exception as err:
            errmsg = ('Error in regex syntax (%s). Remember to use python-'
                      'style regex instead of shell-style.' % err)
        if errmsg:
            raise CannotConfigureRigError(errmsg)
        return message

    def setup(self):
        """
        Watch logs and/or unit files for the provided message
        """
        self.counter = 0
        self.message = self._sanitize_message(self.get_option('message'))
        watch_files = []
        if not self.get_option('no_files'):
            for mfile in self.get_option('logfile').split(','):
                if os.path.isfile(mfile):
                    watch_files.append(mfile)
                else:
                    if (mfile == '/var/log/messages' and not
                            self.get_option('no_journal')):
                        self.log_info(
                            'This system does not have a %s file. '
                            'Ignoring files and only watching journal'
                            % mfile)
                        continue
                    msg = "%s is not a file. Aborting..." % mfile
                    raise CannotConfigureRigError(msg)
        if not self.get_option('no_journal'):
            _j = self.get_option('journal')
            self.add_watcher_thread(target=self.watch_journal, args=_j)
        for wfile in watch_files:
            self.add_watcher_thread(target=self.watch_file, args=wfile)

    def watch_journal(self, journals):
        """
        Watches the journal for new entries and compares them to the provided
        message option

        Note that all journals specified will be monitored via single reader
        rather than one per journal like we (need to) do with files. This
        should reduce system load on polling journald to a minimum.

        Positional arguments
            journals - a list of journal units to filter on. If empty, we are
                       monitoring the full system journal

        Returns
            bool - True when message is matched
        """
        _journs = journals.split(',')
        journ = journal.Reader()
        journ.log_level(journal.LOG_INFO)
        journ.this_boot()
        journ.this_machine()
        self.log_info("Beginning watch of journal(s): %s" % _journs)
        for unit in _journs:
            if unit == 'system':
                continue
            j = unit
            if not j.endswith('.service'):
                j += '.service'
            journ.add_match(_SYSTEMD_UNIT=j)
        journ.seek_tail()
        journ.get_previous()
        _poll = select.poll()
        _journ_fd = journ.fileno()
        _poll_event = journ.get_events()
        _poll.register(_journ_fd, _poll_event)

        while True:
            if _poll.poll(self.get_option('interval')):
                if journ.process() == journal.APPEND:
                    for entry in journ:
                        if self._match_line(entry['MESSAGE'], 'journal'):
                            return True

    def _match_line(self, line, src):
        """
        Helper function that simply looks for a regex match between the given
        line or journal entry and the provided message

        Positional arugments
            line - the line check for a match against the message option
            src  - where the line is coming from, used for logging purposes

        Returns
            bool - did line match the message option *and* is the count option
                   threshold met.
        """
        if re.match(self.message, line):
            self.counter += 1
            self.log_info(
                "Matched user-specified message \"%s\" against line \"%s\""
                " from %s. Message counter at %s of max %s."
                % (self.message, line, src, self.counter,
                   self.get_option('count'))
            )
            return self.counter >= self.get_option('count')
        return False

    def _read_file(self, fileobj):
        """
        Takes an open file object and continually reads from it

        Lifted from David Beazley's Generator Tricks for Python.
        """
        fileobj.seek(0, 2)
        while True:
            line = fileobj.readline()
            if not line:
                self.wait_loop()
                continue
            yield line

    def watch_file(self, filename):
        """
        Watches the provided filename for the given message

        Positional arguments:
            filename - the full path of the file to watch

        Blocks until rig option message is matched in the file

        Returns:
             bool - True when message is matched
        """
        with open(filename, 'r') as wfile:
            self.log_info('Beginning watch of file %s' % filename)
            logs = self._read_file(wfile)
            for line in logs:
                line = line.strip()
                if self._match_line(line, filename):
                    return True



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import argparse
import ast
import json
import logging
import os
import sys
import socket

from logging.handlers import RotatingFileHandler
from rigging.exceptions import *

__version__ = '1.1'


class Rigging():
    """
    Main rig class

    All resources and subcommands are initially handled here. Each of the
    supported resources or subcommands should define argparse options that are
    relevant to it.

    Positional arguments:
        parser - argparse.ArgumentParser object
        args - parsed arguments from the calling rig binary

    """

    def __init__(self, parser, args, supported_rigs=None):
        self.parser = parser
        self.args = args
        self.supported_rigs = supported_rigs

    def _setup_logging(self):
        """Setup logging to /var/log/rig/rig.log"""
        self.logger = logging.getLogger('rig')
        self.logger.setLevel(logging.DEBUG)
        hndlr = RigRotatingFileHandler('/var/log/rig/rig.log')
        hndlr.setFormatter(logging.Formatter(
            '%(asctime)s::%(rig_id)s::%(levelname)s: %(message)s'))
        self.logger.addHandler(hndlr)

        # also print to console (optionally)
        self.console = logging.getLogger('rig_console')
        ui = logging.StreamHandler()
        ui.setFormatter(logging.Formatter('%(message)s'))
        self.console.setLevel(logging.INFO)
        self.console.addHandler(ui)

    def log_error(self, msg):
        self.console.error(msg)

    def log_info(self, msg):
        self.console.info(msg)

    def log_debug(self, msg):
        self.console.debug(msg)

    def log_warn(self, msg):
        self.console.warn(msg)

    def parse_rig_args(self):
        """
        Parse passed args from the cmdline right now, rather than waiting for
        a specific rig to do so
        """
        self.args = vars(self.parser.parse_args())

    def get_id(self):
        """
        Return the value of --id, or raise exception if not provided
        """
        self.parse_rig_args()
        if 'id' in self.args.keys() and self.args['id']:
            return self.args['id']
        print('No rig ID provided, specify a rig with -i/--id')
        raise SystemExit(1)

    def execute(self):
        """
        Based on commandline invocation, setup an appropriate rig or execute a
        subcommand.
        """
        if self.args['subcmd'] == 'list':
            self.list_rigs()
            return 0
        if self.args['subcmd'] == 'info':
            try:
                _rig = RigConnection(self.get_id())
                print(_rig.info())
            except MissingSocketError:
                print("No such rig exists: %s" % self.args['id'])
            return
        self._setup_logging()
        if self.args['subcmd'] == 'destroy':
            return self.destroy_rig(self.get_id())
        if self.args['subcmd'] == 'trigger':
            return self.trigger_rig(self.get_id())
        # load known resource monitors
        if self.args['subcmd'] in self.supported_rigs:
            rig = self.supported_rigs[self.args['subcmd']](self.parser)
            if rig._can_run:
                return rig.execute()
        else:
            self.log_error("Unknown rig type %s provided" %
                           self.args['subcmd'])
            return 1

    def trigger_rig(self, rig_id):
        """
        Trigger a rig right now, rather than waiting for a trigger condition to
        be met.

        Returns
            int - exit code based on success of triggering
                0 - success
                1 - failed to trigger
                2 - invalid command invocation
        """
        if rig_id == '-1':
            self.log_error('Error:  \'trigger\' requires a rig id')
            return 2
        try:
            _rig = RigConnection(rig_id)
            try:
                ret = _rig.trigger()
                if ret['success']:
                    print("Manually triggered rig %s" % rig_id)
                else:
                    print("Failed to trigger rig %s" % rig_id)
                    return 1
            except Exception as err:
                self.log_error("Could not manually trigger rig %s: %s"
                               % (rig_id, err))
                return 1
        except MissingSocketError:
            self.log_error("Non-existing rig %s specified" % rig_id)
            return 1
        return 0

    def destroy_rig(self, rig_id):
        """
        Cleanly shutdown an existing rig with the given id

        Returns
            int - exit code based on destroy behavior
                0 - success
                1 - failed to destroy 1 or more rigs
                2 - invalid command invocation
        """
        if rig_id == '-1':
            self.log_error("Error: 'destroy' requires a rig id or 'all'")
            return 2
        existing_rigs = os.listdir('/var/run/rig')
        if rig_id == 'all':
            rig_id = existing_rigs
        if not isinstance(rig_id, list):
            rig_id = [rig_id]
        for rig in rig_id:
            try:
                if rig in existing_rigs:
                    _rig = RigConnection(rig)
                    ret = _rig.destroy()
                    if ret['success'] and ret['id'] == rig:
                        self.log_info("%s destroyed" % ret['id'])
                    else:
                        self.log_error(ret)

                else:
                    self.log_error("Non-existing rig id(s) provided: %s" % rig)
                    return 1
            except BindSocketError:
                if not self.args['force']:
                    self.log_warn("Could not destroy rig %s, rig is not "
                                  "running." % rig)
                else:
                    os.remove('/var/run/rig/%s' % rig)
        return 0

    def list_rigs(self):
        """
        Lists known rigs.
        Rigs are known by delving into /var/run/rig and querying the available
        sockets.

        This does not return anything, but instead prints directly to console.
        """
        _fmt = "{id:7}{pid:7}{rig_type:8}{watch:30} {trigger:<35} {status:<10}"
        try:
            socks = os.listdir('/var/run/rig')
        except Exception:
            socks = []
        rigs = []
        for sock in socks:
            try:
                rig = RigConnection(sock)
                rigs.append(rig.status())
            except Exception as err:
                rigs.append({
                                'id': sock,
                                'pid': ' ',
                                'rig_type': ' ',
                                'watch': '%s' % err,
                                'trigger': '',
                                'status': 'Unknown'
                            })
        print(_fmt.format(
            id='ID',
            pid='PID',
            rig_type='Type',
            watch='Watching',
            trigger='Trigger',
            status='Status'
        ))
        print('=' * 100)
        for rig in rigs:
            print(_fmt.format(**rig))


class RigConnection():
    """
    Connect to an existing rig's socket, and be able to communicate with the
    rig.
    """

    def __init__(self, socket_name):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        _address = "/var/run/rig/%s" % socket_name
        try:
            self.sock.connect(_address)
        except ConnectionRefusedError:
            raise BindSocketError
        except OSError:
            raise MissingSocketError(_address)

    def _rig_communicate(self, command, extra=''):
        """
        Facilitates communicating with the rig over the socket the rig is
        listening on.

        Messages are sent as json, and at current support the following
        elements:
            command:    the command/function that the rig should run/return
            extra:      Optional, any value that should be passed to the
                        function specified by command

        Returns:
            A dict with the rig's ID, the command that was run, a 'success'
            boolean that indicates if the command was run successfully or not,
            and the resulting output from the command.
        """
        cmd = json.dumps({
            'command': command,
            'extra': extra
        })
        self.sock.settimeout(2)
        self.sock.sendall(cmd.encode())
        try:
            data = self.sock.recv(1024)
        except Exception as err:
            return {'id': None, 'command': 'Unknown', 'success': False}
        return data

    def status(self):
        """
        Query the rig's status.

        Returns
            dict of rig's status information
        """
        ret = json.loads(self._rig_communicate('status').decode())
        if ret['success']:
            return ast.literal_eval(ret['result'])
        raise Exception

    def info(self):
        """
        Query detailed rig information
        """
        ret = json.loads(self._rig_communicate('info').decode())
        if ret['success']:
            return json.dumps(ast.literal_eval(ret['result']), indent=4)
        return ''

    def destroy(self):
        """
        Tell the rig to shutdown cleanly and remove its socket.

        Returns
            str matching the rig ID
        """
        return json.loads(self._rig_communicate('destroy').decode())

    def trigger(self):
        """
        Triggers the rig
        """
        return json.loads(self._rig_communicate('manual_trigger').decode())


class RigRotatingFileHandler(RotatingFileHandler):
    """
    The logging module does not create parent directories for specified log
    files.

    This does, so we use it to create the parent directory.
    """

    def __init__(self, filename, mode='a', maxBytes=1048576, backupCount=5,
                 encoding=None, delay=0):
        try:
            path = os.path.split(filename)[0]
            # will fail on python2, we only support python3
            os.makedirs(path, exist_ok=True)
        except Exception as err:
            print('Could not create logfile at %s: %s' % (path, err))
            sys.exit(1)
        RotatingFileHandler.__init__(self, filename, mode, maxBytes,
                                     backupCount, encoding, delay)



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from rigging.actions import BaseAction


class Kdump(BaseAction):
    """
    Generate a vmcore via kdump

    Note that this action does NOT perform any verification of kdump settings
    or configuration. It is assumed that kdump has been properly tested on the
    node prior to a rig being deployed with this action.
    """

    action_name = 'kdump'
    enabling_opt = 'kdump'
    enabling_opt_desc = 'Generate a vmcore when rig is triggered'
    priority = 10000  # this MUST run last in all cases

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--kdump', action='store_true',
                            help=cls.enabling_opt_desc)
        parser.add_argument('--sysrq', default=None,
                            help='set /proc/sys/kernel/sysrq to this value')
        return parser

    def pre_action(self):
        sysrq = self.get_option('sysrq')
        if sysrq is not None:
            if sysrq == 0:
                self.log_error('Setting /proc/sys/kernel/sysrq to 0 will '
                               'disable kdump, cannot continue.')
                return False
            self.log_info("Setting /proc/sys/kernel/sysrq to %s" % sysrq)
            with open('/proc/sys/kernel/sysrq', 'w') as kern_sysrq:
                try:
                    kern_sysrq.write(sysrq)
                except Exception as err:
                    self.log_error("Failed to set /proc/sys/kernel/sysrq: %s"
                                   % err)
                return False
        return True

    def trigger_action(self):
        self.log_info('Writing \'c\' to /proc/sysrq-trigger - look in your '
                      'configured crash location for a vmcore after reboot')
        with open('/proc/sysrq-trigger', 'w') as sysrq:
            sysrq.write('c')

    def action_info(self):
        return 'A vmcore saved in your configured crash dump location'



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from rigging.rigs import BaseRig
from warnings import catch_warnings, simplefilter

import os
import psutil


class System(BaseRig):
    """
    Monitor overall system utilization metrics.

    These metrics may cover CPU, memory, and/or disk usage.
    """

    parser_description = 'Monitor overall system utilization metrics'
    _cpu_metrics = ['iowait', 'steal', 'nice', 'guest', 'user']
    _memory_metrics = ['available', 'used', 'free', 'slab']

    def set_parser_options(self, parser):
        for metric in self._cpu_metrics:
            _opt = "--%s" % metric
            parser.add_argument(_opt, type=float,
                                help="Threshold for %s" % metric)
        for mem in self._memory_metrics:
            _option = "--%s" % mem
            parser.add_argument(_option, type=int,
                                help="Threshold for %s memory in MiB" % mem)
        parser.add_argument('--cpuperc', type=float,
                            help="Total CPU usage as a percentage")
        parser.add_argument('--memperc', type=float,
                            help="Total memory usage as a percentage")
        parser.add_argument('--loadavg', type=float,
                            help='Single interval load average threshold')
        parser.add_argument('--loadavg-interval', type=int, default=1,
                            choices=[1, 5, 15],
                            help='Which minute interval to watch. Default: 1')
        parser.add_argument('--temp', type=int,
                            help="CPU temperature in degrees Celsius")
        parser.add_argument('--cpu-id', default=0, type=int,
                            help="Physical CPU package ID to monitor temp for")
        return parser

    @property
    def watching(self):
        # TODO: properly define what we're watching
        return 'system utilization'

    @property
    def trigger(self):
        ret = []
        for conf in self.conf:
            if not self.conf[conf]:
                continue
            marker = 'above'
            if conf in self._memory_metrics:
                if conf in ['available', 'free']:
                    marker = 'below'
                _watch = "%s memory" % conf
            elif conf in self._cpu_metrics:
                _watch = "cpu %s" % conf
            ret.append("%s %s %s" % (_watch, marker, self.conf[conf]))
        if self.get_option('cpuperc'):
            ret.append("CPU usage above %s%%" % self.get_option('cpuperc'))
        if self.get_option('memperc'):
            ret.append("Memory usage above %s%%" % self.get_option('memperc'))
        if self.get_option('loadavg'):
            ret.append("System loadavg above %s" % self.get_option('loadavg'))
        if self.get_option('temp'):
            ret.append("CPU temperature above %s" % self.get_option('temp'))
        return ', '.join(r for r in ret)

    def _compile_opts_as_dict(self):
        self.conf = {}
        for arg in self.args:
            if arg in self._cpu_metrics or arg in self._memory_metrics:
                self.conf[arg] = self.args[arg]

    def setup(self):
        self._compile_opts_as_dict()
        # CPU metrics
        cpu = {}
        for _met in self._cpu_metrics:
            if _met in self.conf.keys() and self.conf[_met]:
                cpu[_met] = self.conf[_met]
        mem = {}
        for _mem in self._memory_metrics:
            if _mem in self.conf.keys() and self.conf[_mem]:
                mem[_mem] = self.conf[_mem]
        if self.get_option('memperc'):
            mem['percent'] = self.get_option('memperc')
        if cpu:
            self.add_watcher_thread(self.watch_util_metrics,
                                    args=(cpu, psutil.cpu_times_percent,
                                          'cpu')
                                    )
        if self.get_option('cpuperc'):
            self.add_watcher_thread(self.watch_cpu_utilization,
                                    args=(self.get_option('cpuperc'),))
        if mem:
            self.add_watcher_thread(self.watch_util_metrics,
                                    args=(mem, psutil.virtual_memory,
                                          'memory')
                                    )
        if self.get_option('loadavg'):
            self.add_watcher_thread(self.watch_loadavg,
                                    args=(self.get_option('loadavg')))
        if self.get_option('temp'):
            cpu_id = self._get_package_index(self.get_option('cpu_id'))
            self.add_watcher_thread(self.watch_temp,
                                    args=(self.get_option('temp'),
                                          cpu_id))

    def watch_temp(self, temp, phys_id=0):
        """
        Watch CPU temperature for meeting or exceeding the temperature given.

        By default this will watch CPU Package 0, meaning the first physically
        installed CPU. This may be overridden by the --cpu-id option. Please
        note this refers _only_ to physical CPU packages, and does not support
        individual core monitoring.

        If the given phys_id does not exist, default back to CPU Package 0.

        Positional arguments:
            :param temp:        Degrees celsius to trigger on when met

        Optional arguments:
            :param phys_id:     ID of the physical CPU package to monitor as
                                defined by --cpu-id
        """
        while True:
            cur_temp = psutil.sensors_temperatures()['coretemp'][phys_id]
            if cur_temp.current >= temp:
                self.log_info("CPU temperature is %s C, exceeding threshold "
                              "of %s C" % (cur_temp.current, temp))
                return True
            self.wait_loop()

    def _get_package_index(self, phys_id):
        """
        Find the index in psutil's returned list of temperatures for the given
        physical CPU package ID.
        """
        # squelch runtimewarnings from psutil, which are not important to
        # our purposes here with temperature data
        with catch_warnings():
            simplefilter("ignore", RuntimeWarning)
            _temps = psutil.sensors_temperatures()
        for _temp in _temps['coretemp']:
            # Only consider physical CPUs, not cores
            if 'Package id' not in _temp.label:
                continue
            if _temp.label.split()[-1] == str(phys_id):
                return _temps['coretemp'].index(_temp)
        # If the physical ID does not exist, fallback to package 0, and log
        self.log_warn("Requested CPU package ID %s does not exist. Using CPU "
                      "package 0 instead" % phys_id)
        return 0

    def watch_loadavg(self, threshold):
        """
        Watch overall system loadavg for exceeding value given to the
        --loadavg option.

        Will watch the 1/5/15 minute interval value based on the value of
        --loadavg-interval
        """
        idx = (1, 5, 15).index(self.get_option('loadavg_interval'))
        while True:
            _val = os.getloadavg()[idx]
            if _val >= threshold:
                self.log_info(
                    "System %s-minute loadavg at %s, exceeding threshold of %s"
                    % (self.get_option('loadavg_interval'), _val, threshold)
                )
                return True
            self.wait_loop()

    def watch_cpu_utilization(self, perc):
        """
        Watch the given poller for a threshold above perc.

        Positional arguments:
            perc        percentage value (float) to use as threshold
        """
        # First iteration returns meaningless data
        psutil.cpu_percent()
        while True:
            _val = psutil.cpu_percent(interval=1)
            if _val > perc:
                self.log_info("CPU usage at %s%%, exceeding threshold of %s"
                              % (_val, perc))
                return True

    def watch_util_metrics(self, metrics, poller, resource):
        """
        Watch the given metric from psutil to see if it exceeds the given
        thresold.

        This is passed a dict where the keys are the metrics returned by
        psutil, and the value is the threshold for triggering the rig

        Positional arguments:
            metrics     dict of keys as the metric to watch and the value
                        being our trigger threshold
            poller      method to call to gain stats to compare against
            resource    what kind of resource we're watching
        """
        _watching = ', '.join("{!s}={!r}".format(key, val)
                              for (key, val) in metrics.items())
        self.log_debug("Beginning watch of %s utilization of: %s"
                       % (resource, _watching))
        # some psutil methods are relative to the previous call, so make the
        # first call before looping to set the right context and not trigger
        # on the first poll.
        poller()
        while True:
            _poll = poller()
            for metric in metrics:
                _val = float(getattr(_poll, metric))
                if resource == 'cpu':
                    if _val > metrics[metric]:
                        self.log_info("CPU utilization of %s is %s, exceeding "
                                      "threshold of %s"
                                      % (metric, _val, metrics[metric]))
                        return True
                elif resource == 'memory':
                    if self._check_memory(metric, _val, metrics[metric]):
                        return True
            self.wait_loop()

    def _check_memory(self, metric, val, threshold):
        """
        Helper method to check if the rig should trigger for memory metrics
        based on how that specific metric needs to be compared to current
        stats
        """
        if metric == 'percent':
            if val > threshold:
                self.log_info("Total memory usage is %s%%, exceeding theshold "
                              "of %s" % (val, threshold))
                return True
        elif metric in ['available', 'free']:
            val = int(val / 1024 / 1024)
            if val < threshold:
                self.log_info("%s memory is %s MiB, below threshold of %s MiB"
                              % (metric, val, threshold))
                return True
        elif metric in ['used', 'slab']:
            val = int(val / 1024 / 1024)
            if val > threshold:
                self.log_info("%s memory is %s MiB, exceeding theshold of %s"
                              % (metric, val, threshold))
                return True
        return False



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from rigging.actions import BaseAction
from os.path import basename, isfile
from os import kill

import psutil
import signal


class Gcore(BaseAction):
    """
    Capture a coredump of a running process with gcore (GDB)
    """

    action_name = 'gcore'
    enabling_opt = 'gcore'
    enabling_opt_desc = 'Capture a coredump of a running process'
    priority = 1
    required_binaries = ('gcore',)
    repeatable = True

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--gcore', nargs='?', action='append', default=[],
                            help=cls.enabling_opt_desc)
        parser.add_argument('--all-pids', action='store_true',
                            help=('Execute over all pids found when using '
                                  'process names'))
        parser.add_argument('--freeze', action='store_true',
                            help=('Freeze the process before core dumping, '
                                  'then thaw after completion'))
        return parser

    def _get_pid_from_name(self, pname):
        """
        Find the PID(s) associated with the given process name
        """
        _procs = []
        filt = ['name', 'exe', 'cmdline', 'pid']
        for proc in psutil.process_iter(attrs=filt):
            if (proc.info['name'] == pname or
                    proc.info['exe'] and basename(proc.info['exe']) == pname or
                    proc.info['cmdline'] and proc.info['cmdline'][0] == pname):
                _procs.append(proc.info['pid'])
        if len(_procs) > 1 and not (self.get_option('all') or
                                    self.get_option('all_pids')):
            msg = ("Multiple PIDs found for process '%s', use --all to watch "
                   "all PIDs" % pname)
            self.log_error(msg)
            raise Exception
        return _procs

    def pre_action(self):
        """
        Handle both process names and specific pids in command line args.

        If a PID is supplied, the core name will be core.$pid. If a process
        name is provided, the core name will be core.$name.$pid
        """
        self.pid_list = []
        # the collected list of requested pids, to be filled later
        procs = []
        # the raw commandline value(s) for --gcore
        _pid = self.get_option('gcore')
        for _p in _pid:
            if _p is None:
                if not self.get_option('process'):
                    msg = ('gcore action must be given a pid or process name, '
                           'or be used with the \'process\' rig type for no'
                           'value')
                    self.log_error(msg)
                    return False
                procs.extend(self.get_option('process'))
            else:
                procs.extend(_p.split(','))
        # actually verify the pids derived from the command line options
        for pid in procs:
            try:
                self.pid_list.append((int(pid), ''))
            except ValueError:
                # we were handed a process name, get the pid of it
                proc_pids = self._get_pid_from_name(pid)
                for proc in proc_pids:
                    self.pid_list.append((int(proc), pid))
        msg = ("Determined pid list to collect coredumps for to be: %s"
               % ','.join(str(p[0]) for p in self.pid_list))
        self.log_debug(msg)
        return True

    def freeze_pid(self, pid):
        """
        Send a SIGSTOP to the specified pid
        """
        if self.get_option('freeze'):
            self.log_debug("Freezing pid %s" % pid)
            try:
                kill(pid, signal.SIGSTOP)
                return True
            except Exception as err:
                self.log_error("Could not send SIGSTOP to %s: %s" % (pid, err))
        return False

    def thaw_pid(self, pid):
        """
        Send a SIGCONT to the specified pid
        """
        self.log_debug("Thawing pid %s" % pid)
        try:
            kill(pid, signal.SIGCONT)
        except Exception as err:
            self.log_error("Could not send SIGCONT to %s: %s" % (pid, err))

    def trigger_action(self):
        for pid in self.pid_list:
            _frozen = self.freeze_pid(int(pid[0]))
            _loc = self.tmp_dir + 'core'
            if pid[1]:
                _loc += ".%s" % pid[1]
            if not psutil.pid_exists(int(pid[0])):
                self.log_error("Cannot collect core for pid %s - pid no "
                               "longer exists" % pid[0])
                continue
            if self.repeat_count:
                _loc = _loc.replace('core', "core-%s" % self.repeat_count)
            self.log_debug("Collecting gcore of %s at %s" % (pid[0], _loc))
            try:
                ret = self.exec_cmd("gcore -o %s %s" % (_loc, pid[0]))
                if ret['status'] == 0:
                    if isfile(_loc + '.%s' % pid[0]):
                        fname = _loc + '.%s' % pid[0]
                    else:
                        fname = ret['stdout'].splitlines()[-2].split()[-1]
                    self.add_report_file(fname)
                if _frozen:
                    self.thaw_pid(int(pid[0]))
            except Exception as err:
                self.log_error("Error collecting coredump of %s: %s"
                               % (pid[0], err))
        return True

    def action_info(self,):
        return ("A coredump for each of the following PIDs: %s"
                % ', '.join(str(p[0]) for p in self.pid_list))



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import fnmatch

from pipes import quote
from rigging.actions import BaseAction
from rigging.exceptions import CannotConfigureRigError

SOS_BIN = '/usr/sbin/sosreport --batch'


class SoSReport(BaseAction):

    action_name = 'sosreport'
    enabling_opt = 'sosreport'
    enabling_opt_desc = 'Generate an sos report when triggered'
    priority = 100
    required_binaries = ('sosreport',)
    sos_opts = ('only_plugins', 'skip_plugins', 'enable_plugins',
                'plugin_option')

    def pre_action(self):
        """
        Performs basic sanity checks against any passed --sos-opts and will
        abort rig creation if suspect items like shell code are found
        """
        _filt = ['<', '>', '|', '&', ';']
        for _opt in self.sos_opts:
            _cmd = self.get_option(_opt)
            if not _cmd:
                continue
            if any(f in _cmd for f in _filt):
                raise CannotConfigureRigError(
                    "Potential shell-code found in option %s. Aborting rig "
                    "configuration." % _opt
                )
        self.configure_sos_cmd()
        if self.get_option('initial_sos'):
            self.log_info('Collecting initial baseline sos report, rig '
                          'deployment will complete after report is collected')
            initial_sos_success = self.generate_sosreport(label='initial')
            if not initial_sos_success:
                self.log_error('Failed to generate initial sos report at rig '
                               'deployment.')
                return False
        return True

    def configure_sos_cmd(self):
        """Based on rig options and defaults, determine the sos report command
        to run
        """
        try:
            cmd = "%s --tmp-dir=%s" % (SOS_BIN, self.tmp_dir)
            for _opt in self.sos_opts:
                if self.get_option(_opt):
                    cmd += " --%s %s" % (
                        _opt.replace('_', '-'),
                        quote(self.get_option(_opt))
                    )
            self.sos_cmd = cmd
        except Exception as err:
            self.log_debug("Could not configure sos command: %s" % err)
            raise

    def generate_sosreport(self, label=None):
        """Execute the sos command defined by this rig
        """
        path = 'unknown'
        try:
            cmd = self.sos_cmd
            if label is not None:
                cmd += " --label=%s" % label
            ret = self.exec_cmd(cmd)
            if ret['status'] == 0:
                for line in ret['stdout'].splitlines():
                    if fnmatch.fnmatch(line, '*sosreport-*tar*'):
                        path = line.strip()
                if path == 'unknown':
                    self.log_error('Could not determine path for sos report')
                    self.log_debug(ret['stdout'])
                self.add_report_file(path)
                return True
            else:
                self.log_error("Error during sos report collection: %s"
                               % ret['stderr'] or ret['stdout'])
                return False
        except Exception as err:
            self.log_debug(err)
            return False

    def trigger_action(self):
        self.log_info('Generating sos report from trigger event')
        return self.generate_sosreport()

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--sosreport', action='store_true',
                            help=cls.enabling_opt_desc)
        parser.add_argument('--initial-sos', action='store_true',
                            help="Capture an sos report when rig starts")
        parser.add_argument('-e', '--enable-plugins', type=str,
                            help="Explicitly enable these sos report plugins")
        parser.add_argument('-k', '--plugin-option', type=str,
                            help="Specify sos report plugin options")
        parser.add_argument('-n', '--skip-plugins', type=str,
                            help="Skip these sos report plugins")
        parser.add_argument('-o', '--only-plugins', type=str,
                            help="Only enable these sos report plugins")
        return parser

    def action_info(self):
        return "An sos report from the host in %s" % self.tmp_dir



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import datetime as dt
import glob
import os
import shlex
import subprocess

from pipes import quote
from rigging.actions import BaseAction
from rigging.exceptions import CannotConfigureRigError

TCPDUMP_BIN = '/usr/sbin/tcpdump'
# -Z is needed to avoid the privilege drop that happens before opening the
# first savefile, which would result in an ENOPERM and a failed rig
TCPDUMP_OPTS = '-Z root -n'


class Tcpdump(BaseAction):
    """
    Start a tcpdump and stop it once the trigger condition is met
    """

    action_name = 'tcpdump'
    enabling_opt = 'tcpdump'
    enabling_opt_desc = 'Start a tcpdump that ends when rig is triggered'
    priority = 2
    required_binaries = ('tcpdump',)

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--tcpdump', action='store_true',
                            help=cls.enabling_opt_desc)
        parser.add_argument('--filter', default=None,
                            help='Packet filter to use')
        parser.add_argument('--iface', '--interface', default='eth0',
                            help='Interface to listen on (default eth0)')
        parser.add_argument('--dump-size', default=10, type=int,
                            help='Maximum size of packet capture in MB')
        parser.add_argument('--captures', default=1, type=int,
                            help='Number of capture files to keep')
        parser.add_argument('--snaplen', '--snapshot-length', default=0,
                            type=int, dest='snaplen',
                            help='Snapshot length of packets captured')
        return parser

    def pre_action(self):
        """
        Launch the tcpdump
        """
        _date = dt.datetime.today().strftime("%d-%m-%Y-%H:%M:%S")
        name = "%s-%s-%s" % (self.exec_cmd('hostname')['stdout'].strip(),
                             _date,
                             self.get_option('iface'))
        self.loc = "%s%s.pcap" % (self.tmp_dir, name)
        cmd = ("%s %s -i %s -s %s -C %s -W %s "
               % (TCPDUMP_BIN, TCPDUMP_OPTS, self.get_option('iface'),
                  self.get_option('snaplen'), self.get_option('dump_size'),
                  self.get_option('captures'))
               )
        cmd += "-w %s" % self.loc
        if self.get_option('filter'):
            cmd += " %s" % quote(self.get_option('filter'))
        self.log_debug("Running tcpdump as '%s'" % cmd)
        self.devnull = open(os.devnull, 'w')
        self.proc = subprocess.Popen(shlex.split(cmd), shell=False,
                                     stdout=self.devnull,
                                     stderr=subprocess.PIPE)
        try:
            # if we hit an error in the first second of execution, it means
            # tcpdump was configured incorrectly
            stdout, stderr = self.proc.communicate(timeout=1)
            if stderr:
                raise CannotConfigureRigError(stderr.decode('utf-8').strip())
        except subprocess.TimeoutExpired:
            pass
        self.log_debug("Started background tcpdump on interface '%s'"
                       % self.args['iface'])
        return True

    def trigger_action(self):
        self.log_debug("Stopping tcpdump")
        try:
            self.proc.terminate()
            _files = glob.glob(self.loc + '*')
            self.add_report_file(_files)
            self.devnull.close()
        except Exception as err:
            self.log_error("Could not stop tcpdump: %s" % err)
        return True

    def cleanup(self):
        try:
            self.proc.terminate()
            self.devnull.close()
        except Exception:
            pass

    def action_info(self):
        return "A packet capture from interface %s using filter %s" % (
            self.get_option('iface'),
            self.get_option('filter')
        )



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from rigging.actions import BaseAction


class Noop(BaseAction):
    """
    Do nothing. Used for testing rig configurations.
    """

    action_name = 'noop'
    enabling_opt = 'noop'
    enabling_opt_desc = 'Do nothing. Use for testing rig configurations'
    repeatable = True

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--noop', action='store_true',
                            help=cls.enabling_opt_desc)
        return parser

    def trigger_action(self):
        self.log_info('No-op action triggered. Doing nothing.')
        return True

    def action_info(self):
        return 'This action will generate no content'



# Copyright (C) 2022 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

from datetime import datetime, timedelta
from rigging.actions import BaseAction
from subprocess import Popen, PIPE, STDOUT
from threading import Thread

import shlex
import time
import os


class BackgroundCollector(Thread):
    """Used by this action to run a thread in the background that will
    periodically run a command or collect a file's content.
    """

    def __init__(self, interval):
        self.interval = interval
        self._running = False
        self.stopped = False
        super(BackgroundCollector, self).__init__(daemon=True)

    def start(self):
        self._running = True
        super(BackgroundCollector, self).start()

    def stop(self):
        self._running = False

    def _continue(self):
        """Block for up to self.interval seconds, before returning True to
        allow the run() loop to continue on to the next iteration. If during
        this time self._running is set to False (e.g. via self.stop()), then
        immediately return False to break out of the run loop without having
        to wait for the entire interval to pass

        We need to use this approach instead of a threading.Event() pill
        because of the fact that we double fork the rig's main process in order
        to run in the background.

        :returns:   True if run() should continue executing, else false
        :rtype:     ``bool``
        """
        _max = datetime.now() + timedelta(seconds=self.interval)
        while datetime.now() < _max:
            if not self._running:
                return False
            time.sleep(1)
        return True

    def run(self):
        # get the first set of data when the rig starts, not after the first
        # interval has passed
        self._run_and_write()
        while self._continue():
            self._run_and_write()

        self.stopped = True

    def _write_with_header(self, fobj, content):
        """
        Write our timestamped header to the output file followed by whatever
        has been collected by the background thread.

        :param fobj:    The open file object to write the header to
        """
        try:
            now = datetime.now()
            fobj.write("==== %s ====\n" % now)
            fobj.write(content + '\n')
        except Exception as err:
            fobj.write("Rig error: Unable to write content: %s" % err)

    def _run_and_write(self):
        """
        What is actually executed in the background while the thread is
        running.

        Will be overridden by the actual collector being implemented.
        """
        raise NotImplementedError


class FileCollector(BackgroundCollector):
    """
    Used to collect the contents of a particular file as defined by this rig.
    """

    def __init__(self, to_copy, fname, interval):
        """
        :param to_copy:     The name of the file to capture
        :type to_copy:      ``str``

        :param fname:       Path to save the file contents to, including tmpdir
        :type fname:        ``str``

        :param interval:    Time to wait between subsequent collections
        :type interval:     ``int``
        """
        self.to_copy = to_copy
        self.fname = fname
        super(FileCollector, self).__init__(interval=interval)

    def _run_and_write(self):
        with open(self.to_copy, 'r') as _to_copy:
            with open(self.fname, 'a') as out_file:
                self._write_with_header(out_file, _to_copy.read())


class CmdCollector(BackgroundCollector):
    """
    This class is used to control the execution and output storing of a command
    used by this rig. Each command that we'll need to repeatedly capture will
    be instantiated via a CmdMonitor() and run in a background thread.
    """

    def __init__(self, cmd, fname, interval):
        """
        :param cmd:     The command the execute and collect the output of
        :type cmd:      ``str``

        :param fname:   Where to save the output to, including rig's tmpdir
        :type fname:    ``str``

        :param stopper: An event to watch for when the thread should exit
        :type stopper:  ``threading.Event``

        :param interval:    Time to wait between subsequent ``cmd`` executions
        :type interval:     ``int``
        """
        self.cmd = shlex.split(cmd)
        self.fname = fname
        super(CmdCollector, self).__init__(interval=interval)

    def _run_and_write(self):
        """
        Run the monitor's command and write the output to requested file
        """
        with open(self.fname, 'a') as ofile:
            _proc = Popen(self.cmd, shell=False, stdout=PIPE,
                          stderr=STDOUT, encoding='utf-8')
            try:
                timeout = max(1, self.interval / 2)
                msg, serr = _proc.communicate(timeout=timeout)
            except Exception as err:
                _proc.terminate()
                msg = "Could not collect command output: %s" % err
            self._write_with_header(ofile, msg)


class SysMonitor(BaseAction):
    """
    Record various system stats over the course of the rig's lifetime until
    the rig triggers or is destroyed.

    This is inspired by the `monitor.sh` script used by Red Hat Support
    engineers typically in conjunction with networking related issues.

    See https://access.redhat.com/articles/1311173#monitorsh-script-notes-3
    for context.

    The --disable-monitor-defaults option toggles a standard set of monitors
    that are largely based on the script above.
    """

    action_name = 'monitor'
    enabling_opt = 'monitor'
    enabling_opt_desc = ('record various system statistics over the lifetime '
                         'of the rig')

    @classmethod
    def add_action_options(cls, parser):
        parser.add_argument('--monitor', action='store_true',
                            help=cls.enabling_opt_desc)
        parser.add_argument('--monitor-files', default=[], action='rigextend',
                            help='List of files to periodically monitor')
        parser.add_argument('--monitor-commands', default=[],
                            action='rigextend',
                            help='List of commands to periodically execute')
        parser.add_argument('--disable-monitor-defaults', action='store_true',
                            help='Disable the default resource monitors')

    def pre_action(self):
        """
        Launch the background threads to perform monitoring
        """
        self.procs = {}
        _cmdmons = []
        _filemons = []

        for cmd in self.get_option('monitor_commands'):
            _cmd = cmd.split()
            # tuples used to create the CmdCollectors.
            # Form of (executable, options)
            _cmdmons.append((_cmd[0], ' '.join(_cmd[1:])))

        for _fname in self.get_option('monitor_files'):
            _filemons.append(_fname)

        if not self.get_option('disable_monitor_defaults'):
            _cmdmons.extend([
                ('netstat', '-s'),
                ('nstat', '-az'),
                ('ss', '-noemitaup'),
                ('ps', '-alfe'),
                ('top', '-c -b -n 1'),
                ('numastat', ''),
                ('ip', 'neigh show'),
                ('tc', '-s qdisc')
            ])

            mdevs = self.exec_cmd('tc qdisc show')
            if mdevs['status'] == 0:
                for mdevln in mdevs['stdout'].splitlines():
                    if 'qdisc mq' in mdevln:
                        mdev = mdevln.split('dev')[1].strip().split()[0]
                        opts = "-s class show dev %s" % mdev
                        _cmdmons.append(('tc', opts))

            _filemons.extend([
                '/proc/interrupts',
                '/proc/vmstat',
                '/proc/net/softnet_stat',
                '/proc/softirqs',
                '/proc/net/sockstat',
                '/proc/net/sockstat6',
                ('/proc/net/dev', 'netdev'),
                ('/proc/net/sctp/assocs', 'sctp_assocs'),
                ('/proc/net/sctp/snmp', 'sctp_snmp')
            ])

        for cmd in _cmdmons:
            if self._check_exists(cmd[0]):
                cmdfn = "%s_%s" % (cmd[0],
                                   cmd[1].replace(' ', '_').replace('/', '.'))
                _out = os.path.join(self.tmp_dir, cmdfn)
                _cmdstr = ' '.join(c for c in cmd)
                self.procs[cmdfn] = CmdCollector(_cmdstr, _out,
                                                 self.get_option('interval'))
            else:
                fullcmd = "%s %s" % (cmd[0], cmd[1])
                self.log_info(
                    "Unable to create command monitor for '%s': '%s' not found"
                    % (fullcmd, cmd[0])
                )

        for _file in _filemons:
            if isinstance(_file, tuple):
                _src = _file[0]
                _fname = _file[1]
            else:
                _src = _file
                _fname = _file.split('/')[-1]
            if os.path.exists(_src):
                _dest = os.path.join(self.tmp_dir, _fname)
                self.procs[_fname] = FileCollector(_src, _dest,
                                                   self.get_option('interval'))
            else:
                self.log_info(
                    "Unable to create file monitor for '%s': no such file"
                    % _src
                )

        for proc in self.procs:
            self.log_debug("Starting '%s' periodic collector" % proc)
            self.procs[proc].start()
        return True

    def trigger_action(self):
        # send the stop signal to all monitors
        for proc in self.procs:
            self.procs[proc].stop()
            self.add_report_file(self.procs[proc].fname)
        while not all([p.stopped for p in self.procs.values()]):
            self.log_info(
                "Waiting for collectors %s to stop"
                % (', '.join(p for p in self.procs if not
                             self.procs[p].stopped))
            )
            time.sleep(1)

    def action_info(self):
        return "A collection of file and/or command outputs taken periodically"



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.


from rigging.rigs import BaseRig
from rigging.exceptions import CannotConfigureRigError
from os.path import basename

import psutil
import time


class Process(BaseRig):
    """
    Watch a process for state changes and trigger when user-provided definition
    is met.

    This definition may be a process state (such as D-state) or process
    resource consumption such as CPU or memory.

    The following parameters are available as rig options:
        :opt proc:    The PID or name of the process to monitor
        :opt state:   The status of a given process on which to trigger
        :opt rss:     The resident set size threshold on which to trigger
        :opt vms:     The virtual memory size threshold on which to trigger
        :opt memperc: The percentage of memory usage on which to trigger
        :opt cpuperc: The percentage of CPU usage on which to trigger

    """

    parser_description = ('Monitor a process for state or resource consumption'
                          ' thresholds.')
    rig_wide_opts = ('process', 'all')

    def set_parser_options(self, subparser):
        subparser.add_argument('-p', '--process', action='append',
                               help='PID or name of process to watch')
        subparser.add_argument('--all', action='store_true', default=False,
                               help='Watch all PIDs for a process name')
        subparser.add_argument('--state',
                               help='What process status to trigger on')
        subparser.add_argument('--rss',
                               help='Memory RSS threshold to trigger on')
        subparser.add_argument('--vms',
                               help=('Virtual Memory Size threshold to '
                                     'trigger on'))
        subparser.add_argument('--memperc',
                               help=('Percentage of system memory usage to '
                                     'trigger on'))
        subparser.add_argument('--cpuperc',
                               help='Percentage of CPU usage to trigger on')
        return subparser

    @property
    def watching(self):
        return ', '.join(str(p) for p in self.proc_list)

    @property
    def trigger(self):
        triggers = ''
        if self.get_option('state'):
            triggers += 'State: %s ' % self.get_option('state')
        if self.get_option('rss'):
            triggers += "RSS usage above %s " % self.get_option('rss')
        if self.get_option('vms'):
            triggers += "VMS usage above %s " % self.get_option('vms')
        if self.get_option('memperc'):
            triggers += "%%Mem usage above %s%% " % self.get_option('memperc')
        if self.get_option('cpuperc'):
            triggers += "%%CPU usage above %s%%" % self.get_option('cpuperc')
        return triggers

    def _get_pid_from_name(self, pname):
        """
        Find the PID(s) associated with the given process name
        """
        _procs = []
        filt = ['name', 'exe', 'cmdline', 'pid']
        for proc in psutil.process_iter(attrs=filt):
            if (proc.info['name'] == pname or
                    proc.info['exe'] and basename(proc.info['exe']) == pname or
                    proc.info['cmdline'] and proc.info['cmdline'][0] == pname):
                _procs.append(proc.info['pid'])
        if len(_procs) > 1 and not self.get_option('all'):
            msg = ("Multiple PIDs found for process '%s', use --all to watch "
                   "all PIDs" % pname)
            raise CannotConfigureRigError(msg)
        return _procs

    def _get_bytes(self, val):
        """
        Converts the user provided string to an integer representing the
        memory threshold specified in bytes.
        """
        suffixes = [('K', 10), ('M', 20), ('G', 30)]
        if not any(suf[0] in val for suf in suffixes):
            raise CannotConfigureRigError(
                "Values must be given with K, M, or G suffixes"
            )
        for suff in suffixes:
            if suff[0].lower() in val.lower():
                _suf = suff
        return int(val.split(_suf[0])[0]) << _suf[1]

    def _validate(self):
        """
        This will do several tasks that ensure the rig is properly defined:

        First, it will validate the given PID, or convert a process name
        to a PID or list of PIDS

        Second, it will validate that all given or discovered PIDs exist

        Third, it will validate the provided status is one we can match with
        psutil. Or it will determine the memory size in bytes to use as our
        threshold.
        """

        # make sure we're actually watching something
        if not any(self.get_option(arg) for arg in
                   ['rss', 'vms', 'state', 'memperc', 'cpuperc']):
            raise CannotConfigureRigError("No metric given to watch. See "
                                          "'rig process --help'")
        self.proc_list = []
        procs = []
        _procs = self.get_option('process')
        for proc in _procs:
            procs.extend(proc.split(','))
        for process in procs:
            try:
                _proc = int(process)
                self.proc_list.append(_proc)
            except ValueError:
                _proc = self._get_pid_from_name(process)
                if not _proc:
                    raise CannotConfigureRigError("No processes found for '%s'"
                                                  % process)
                self.proc_list.extend(_proc)
        if not self.proc_list:
            raise CannotConfigureRigError('No valid PIDs provided. Aborting.')
        for proc in self.proc_list:
            if psutil.pid_exists(int(proc)):
                continue
            # if any PIDs don't actually exist, abort.
            raise CannotConfigureRigError("Invalid PID provided: %s" % proc)
        if self.get_option('state'):
            _state_map = {
                psutil.STATUS_RUNNING: ('R', 'run'),
                psutil.STATUS_SLEEPING: ('S', 'sleep'),
                psutil.STATUS_DISK_SLEEP: ('D', 'UN'),
                psutil.STATUS_STOPPED: ('T', 'stop'),
                psutil.STATUS_ZOMBIE: ('Z', 'zomb'),
                psutil.STATUS_DEAD: ('X',)
            }
            _state = self.get_option('state').strip('!')
            for key, value in _state_map.items():
                if _state in value:
                    _repl = self.get_option('state').replace(_state, key)
                    self.set_option('state', _repl)
            if self.get_option('state').strip('!') not in _state_map.keys():
                msg = ("Invalid status '%s' provided. See 'man rig' for a "
                       "full list of supported statuses."
                       % self.get_option('state'))
                raise CannotConfigureRigError(msg)
        if self.get_option('rss'):
            self.rss_limit = self._get_bytes(self.get_option('rss'))
        if self.get_option('vms'):
            self.vms_limit = self._get_bytes(self.get_option('vms'))
        self.set_option('process', self.proc_list)
        return True

    @property
    def all_pids_are_dead(self):
        return all([not psutil.pid_exists(p) for p in self.proc_list])

    def _hold_thread(self, process, resource):
        """
        When a monitor thread is about to end due to a pid no longer existing,
        instead block that thread until all pids that the rig is monitoring
        have exited as well
        """
        while not self.all_pids_are_dead:
            time.sleep(self.get_option('interval'))
        self.log_info("%s monitor thread for pid %s exiting."
                      % (resource, process))
        return False

    def setup(self):
        """
        Create a monitoring thread for each PID discovered or provided for each
        process item we can monitor for and the user requested.
        """
        self._validate()
        stat = self.get_option('state')
        for proc in self.proc_list:
            if stat:
                invert = stat.startswith('!')
                self.add_watcher_thread(self.watch_process_for_status,
                                        args=(proc, stat, invert))
            if self.get_option('rss'):
                self.add_watcher_thread(self.watch_process_for_mem,
                                        args=(proc, self.rss_limit, 'rss'))
            if self.get_option('vms'):
                self.add_watcher_thread(self.watch_process_for_mem,
                                        args=(proc, self.vms_limit, 'vms'))
            if self.get_option('memperc'):
                limit = float(self.get_option('memperc'))
                self.add_watcher_thread(self.watch_process_for_perc,
                                        args=(proc, limit, 'memory'))
            if self.get_option('cpuperc'):
                limit = float(self.get_option('cpuperc'))
                self.add_watcher_thread(self.watch_process_for_perc,
                                        args=(proc, limit, 'cpu'))

    def watch_process_for_status(self, process, status, invert):
        """
        Used to watch the given process(es) for status changes and compares
        the current status to the trigger defined by the status option.

        Positional arguments
            process         The PID of the process to watch
            status          The state of the process to trigger on
            invert          Should the boolean comparison for status against
                            current actual process state be reversed
        """
        proc = psutil.Process(process)
        user_stat = status.strip('!')
        self.log_info("Beginning watch of process %s for status '%s'"
                      % (process, status))
        while True:
            try:
                _stat = proc.status()
                if status == '!running' and _stat == 'sleeping':
                    pass
                elif ((_stat == user_stat and invert is False) or
                      (_stat != user_stat and invert is True)):
                    self.log_info("State for process %s is '%s', matching "
                                  "trigger state '%s'."
                                  % (process, _stat, status))
                    return True
                self.wait_loop()
            except psutil.NoSuchProcess:
                if status == '!running':
                    self.log_info("Process %s is no longer running, matching "
                                  "trigger state '%s'" % (process, status))
                    return True
                self.log_info("Process %s is no longer running and desired "
                              "trigger state is not '!running'. Stopping "
                              "watcher thread without triggering rig."
                              % process)
                return False

    def watch_process_for_mem(self, process, limit, mem_type):
        """
        Watch the given process(es) for mem consumption values that exceed the
        amount specified by limit.

        Positional arguments:
            process     The PID of the process to watch
            limit       The usage threshold in bytes
            mem_type    The memory stat to watch - e.g. rss or vms.
        """
        proc = psutil.Process(process)
        self.log_info("Beginning watch of process %s for %s memory usage of "
                      "%s bytes or higher" % (process, mem_type, limit))
        while True:
            try:
                _mem = proc.memory_info()
                _check = getattr(_mem, mem_type)
                if _check >= limit:
                    self.log_info("Process %s has %s usage of %s, exceeding "
                                  "trigger threshold of %s."
                                  % (process, mem_type, _check, limit))
                    return True
                self.wait_loop()
            except psutil.NoSuchProcess:
                self.log_info("Process %s is no longer running." % process)
                return self._hold_thread(process, "%s consumption" % mem_type)

    def watch_process_for_perc(self, process, limit, resource):
        """
        Watch the given process(es) for memory consumption higher than the
        percentage of system memory defined by limit.

        Positional arguments:
            process     The PID of the process to watch
            limit       The usage percentage threshold as a float
            resource    'memory' or 'cpu' - which to monitor
        """
        proc = psutil.Process(process)
        self.log_info("Beginning watch of process %s for total %s usage of"
                      " %s%% or higher" % (process, resource, limit))
        while True:
            try:
                if resource == 'memory':
                    _perc = float("{0:.2f}".format(proc.memory_percent()))
                elif resource == 'cpu':
                    _perc = float(proc.cpu_percent())
                if _perc >= limit:
                    self.log_info("Process %s has total %s usage of %s%% "
                                  "exceeding trigger threshold of %s%%"
                                  % (process, resource, _perc, limit))
                    return True
                self.wait_loop()
            except psutil.NoSuchProcess:
                self.log_info("Process %s is no longer running." % process)
                return self._hold_thread(process, resource)



from setuptools import setup, find_packages
from rigging import __version__ as VERSION

setup(
    name='rig',
    version=VERSION,
    description='Monitor a system for events and trigger specific actions',
    long_description=("Rig is a utility designed to watch or monitor specific "
                      "system resources (e.g. log files, journals, network "
                      "activity, etc...) and then take specific action when "
                      "the trigger condition is met. Its primary aim is to "
                      "assist in troubleshooting and data colelction for "
                      "randomly occurring events."),
    author='Jake Hunsaker',
    author_email='jhunsake@redhat.com',
    license='GPLv2',
    url='https://github.com/TurboTurtle/rig',
    classifiers=[
                'Intended Audience :: System Administrators',
                'Topic :: System :: Systems Administration',
                ('License :: OSI Approved :: GNU General Public License v2 '
                 "(GPLv2)"),
                ],
    packages=find_packages(),
    install_requires=['psutil > 5', 'systemd-python'],
    scripts=['rig'],
    data_files=[
        ('share/licenses/rig', ['LICENSE']),
        ('share/man/man1/', ['man/en/rig.1'])
    ])



# Copyright (C) 2019 Red Hat, Inc., Jake Hunsaker <jhunsake@redhat.com>

# This file is part of the rig project: https://github.com/TurboTurtle/rig
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# version 2 of the GNU General Public License.
#
# See the LICENSE file in the source distribution for further information.

import re
import shlex

from rigging.rigs import BaseRig
from rigging.exceptions import CannotConfigureRigError
from subprocess import Popen, PIPE, TimeoutExpired

PING_BIN = '/usr/bin/ping'


class Ping(BaseRig):
    """
    A basic connectivity testing rig. While running, this rig will send a ping
    at a defined interval and record the results. After the ping completes,
    the recorded results are compared to a defined trigger condition (e.g.
    number of packets lost).

    Resource options:
        :opt host:              The IP or hostname to ping
        :opt ping-timeout:      Timeout for individual pings
        :opt ping-time-ms       Max time for ping response in ms
        :opt ping-interval:     Time to wait between pings
        :opt lost-count:        Number of packets to lose to trigger rig
    """

    parser_description = 'Send pings at a regular interval'

    def set_parser_options(self, subparser):
        subparser.add_argument('--host',
                               help='IP address or hostname to ping')
        subparser.add_argument('--ping-timeout', default=1, type=float,
                               help='Timeout in seconds for response')
        subparser.add_argument('--lost-count', default=1, type=int,
                               help='Packet loss count threshold for trigger')
        subparser.add_argument('--ping-ms-max', default=None, type=int,
                               help='Max RTT threshold for pings')
        subparser.add_argument('--ping-ms-count', default=5, type=int,
                               help='Threshold for packets over max RTT')
        return subparser

    @property
    def watching(self):
        return "Pinging %s" % self.get_option('host')

    @property
    def trigger(self):
        msg = "%s lost/dropped packets" % self.get_option('lost_count')
        if self.get_option('ping_ms_max'):
            msg += (" or %s packets exceeding RTT of %s ms"
                    % (self.get_option('ping_ms_count'),
                       self.get_option('ping_ms_max'))
                    )
        return msg

    def reset_counters(self):
        self.lost_packets = 0
        self.packet_ms_count = 0

    def _run_ping(self):
        try:
            proc = Popen(shlex.split(self.ping_cmd), stderr=PIPE, stdout=PIPE)
            stdout, stderr = proc.communicate(
                                timeout=self.get_option('ping_timeout'))
            rc = proc.returncode
        except TimeoutExpired:
            # need to set this manually
            self.log_info("Registered packet timeout to %s"
                          % self.get_option('host'))
            self.lost_packets += 1
            # a timed-out packet is by nature exceeding the RTT
            self.packet_ms_count += 1
            if self._check_loss():
                return False
            else:
                return True
        if rc == 0 or rc == 1:
            if self.parse_ping_failure(stdout):
                return False
            return True
        elif rc == 2:
            msg = "Host %s did not resolve" % self.get_option('host')
        else:
            msg = "Error running ping. Return code %s" % rc
        raise Exception(msg)

    def parse_ping_failure(self, output):
        status = output.splitlines()[-2].decode('utf-8', 'ignore')
        recv = int(re.match(r'.*, (\d+) received, .*', status).group(1))
        if recv == 0:
            self.log_debug("Registered lost packet to %s"
                           % self.get_option('host'))
            self.lost_packets += 1
            if self._check_loss():
                self.log_info("Number of lost packets is %s, triggering "
                              "rig" % self.lost_packets)
                return True
        ms_status = output.splitlines()[-1].decode('utf-8', 'ignore')
        if self.get_option('ping_ms_max') and ms_status:
            # we have to match on the summary line at the end, as it is more
            # reliable for standardized output than the individual ping result
            # lines returned during ping execution
            ms_reg = r'(.* =) ((\d+.\d+)/(\d+.\d+)/(\d+.\d+)/(.*) ms)'
            ms = re.match(ms_reg, ms_status)
            if ms:
                ms = int(float(ms.group(5)))
            if ms >= int(self.get_option('ping_ms_max')):
                self.log_debug("Registered packet with RTT of %s (threshold: "
                               "%s)" % (ms, self.get_option('ping_ms_max')))
                self.packet_ms_count += 1
                if self.packet_ms_count >= self.get_option('ping_ms_count'):
                    self.log_info("Number of packets exceeding RTT threshold "
                                  "exceeded, triggering rig.")
                    return True
        return False

    def _test_initial_ping(self):
        """
        Perform a sanity check ping before launching the rig.
        """
        self.log_debug('Sending a sanity check ping before initialization')
        try:
            self._run_ping()
            if self.lost_packets == 0:
                self.log_debug('Sanity check passed. Ping successful')
                return True
            else:
                raise CannotConfigureRigError(
                    "Initial ping response not received from %s"
                    % self.get_option('host')
                )
        except CannotConfigureRigError:
            raise
        except Exception as err:
            raise CannotConfigureRigError(err)

    def setup(self):
        if not self.get_option('host'):
            raise CannotConfigureRigError('Target host must be provided')
        self.lost_packets = 0
        self.packet_ms_count = 0
        if (self.get_option('ping_ms_max') and self.get_option('ping_timeout')
                < (self.get_option('ping_ms_max') / 1000)):
            _new_timeout = (self.get_option('ping_timeout') +
                            (self.get_option('ping_ms_max') / 1000))
            self.log_info("Ping timeout lower than configured RTT threshold. "
                          "Increasing ping timeout to %s seconds."
                          % _new_timeout)
            self.set_option('ping_timeout', _new_timeout)
        self.ping_cmd = "%s -c 1 %s" % (PING_BIN, self.get_option('host'))
        self._test_initial_ping()
        self.add_watcher_thread(target=self.ping_host, args=None)

    def ping_host(self, hldr):
        """
        Actually ping the host

        Positional arguments:
            hldr:       A holder to account for start_watcher_threads() always
                        passing a positional arg. Unused.
        """
        while True:
            try:
                resp = self._run_ping()
                if not resp:
                    return True
                self.wait_loop()
            except Exception as err:
                self.log_error(err)
                # don't inadvertantly trigger rig on an error
                return False

    def _check_loss(self):
        return self.lost_packets >= self.get_option('lost_count')
